"""Functions to send statistics to ONT."""
import configparser
import platform
import socket
import uuid

import requests

from epi2melabs import CONTAINER_META

ENDPOINT = 'https://ping.oxfordnanoportal.com/epilaby'


def _send_ping(data, session, hostname=None, opsys=None):
    """Attempt to send a ping to home.

    :param data: a dictionary containing the data to send (should be
        json serializable).

    :returns: status code of HTTP request. Returns 503 if requests library
        throws an exception.
    """
    if not isinstance(session, uuid.UUID):
        raise ValueError('`session` should be a uuid.UUID object')
    ping_version = '1.1.0'
    if hostname is None:
        hostname = socket.gethostname()
    if opsys is None:
        opsys = platform.platform()
    ping = {
        "tracking_id": {"msg_id": str(uuid.uuid4()), "version": ping_version},
        "hostname": hostname, "os": opsys,
        "session": str(session)}
    ping.update(data)
    try:
        r = requests.post(ENDPOINT, json=ping)
    except Exception:
        return 503  # "Unavailable"
    else:
        return r.status_code


class Pingu(object):
    """Manage the sending of multiple pings."""

    def __init__(
        self, session=None, hostname=None, opsys=None,
        enabled=True, config=CONTAINER_META
    ):
        """Initialize pinger.

        :param session: a UUID session identifier.
        :param enabled: if disabled methods sending pings will not send
            data but rather return it to the caller. When enabled the
            HTTP request response code is returned. This setting is
            overridden by the contents of the config file.
        :param config: a config (ini) file. Settings in the file will
            override values from the environment and keyword arguments.
        """
        if session is None:
            session = uuid.uuid4()
        else:
            if not isinstance(session, uuid.UUID):
                raise ValueError('`session` should be a uuid.UUID object')
        self.session = session
        self.enabled = enabled
        self.hostname = hostname or socket.gethostname()
        self.opsys = opsys or platform.platform()
        self.state = None
        try:
            config = configparser.ConfigParser()
            config.read(CONTAINER_META)
            self.enabled = config['Pings'].getboolean('enabled')
            self.hostname = config['Host']['hostname']
            self.opsys = config['Host']['operating_system']
        except Exception:
            pass

    def send_container_ping(
            self, action, container_stats, image_name, message=None):
        """Ping a status message of a container.

        :param action: one of 'start', 'stop', or 'update'.
        :param container: the output of `container.stats(stream=False)`
        :param image_tag: the name of the image associated with the container.

        :returns: status code of HTTP request.
        """
        allowed_status = {"start", "stop", "update"}
        if action not in allowed_status:
            raise ValueError(
                "`action` was not an allowed value, got: '{}'".format(action))
        ping_data = {
            "data": {
                "source": "container",
                "action": action,
                "container_data": container_stats,
                "image_data": image_name,
                "message": message},
            "session": self.session, "hostname": self.hostname,
            "opsys": self.opsys}
        if self.enabled:
            return _send_ping(**ping_data)
        else:
            return ping_data

    def send_notebook_ping(self, action, notebook, message=None):
        """Ping a message from a notebook.

        :param action: one of 'start', 'stop', or 'update'.
        :param notebook: name of notebook.
        :param message: optional message (must be json serializable).
        """
        allowed_status = {"start", "stop", "update"}
        if action not in allowed_status:
            raise ValueError(
                "`action` was not an allowed value, got: '{}'".format(action))
        # check for spam
        if any((
                action == self.state,
                action == "stop" and self.state is None)):
            return
        self.state = action
        ping_data = {
            "data": {
                "source": "notebook",
                "action": action,
                "notebook_name": notebook,
                "message": message},
            "session": self.session, "hostname": self.hostname,
            "opsys": self.opsys}
        if self.enabled:
            return _send_ping(**ping_data)
        else:
            return ping_data

    def send_workflow_ping(
            self, workflow, message=None, **kwargs):
        """Ping a status message from a workflow.

        :param workflow: name of the workflow being run.
        :param message: optional message (must be json serializable).
        :param kwargs: optional additional data (must be json serializable).

        :returns: status code of HTTP request.
        """
        ping_data = {
            "data": {
                "source": "workflow",
                "workflow": workflow,
                "message": message,
                **kwargs},
            "session": self.session, "hostname": self.hostname,
            "opsys": self.opsys}
        if self.enabled:
            return _send_ping(**ping_data)
        else:
            return ping_data



"""Select helpers for notebooks."""

import functools
import sys

from colorama import Fore, Style
try:
    # we don't want the whole package to depend on these but for
    # functionality that is inherently notebookish these will
    # be present in that context
    from IPython import display
    import ipywidgets as widgets
except ImportError:
    pass


def cecho():
    """Print formatted text.

    Entrypoint to print formatted text:

        cecho error This message is printed in red

    Valid formats are: 'error', 'warning', 'ok', 'success'
    """
    style, *text = sys.argv[1:]
    text = ' '.join(text)
    if style == 'error':
        print(error(text))
    elif style == 'warning':
        print(warning(text))
    elif style == 'ok':
        print(ok(text))
    elif style == 'success':
        print(success(text))
    else:
        print(text)


def error(text):
    """Format text as heavy red."""
    return Fore.RED + Style.BRIGHT + text + Style.RESET_ALL


def warning(text):
    """Format text as heavy magenta."""
    return Fore.MAGENTA + Style.BRIGHT + text + Style.RESET_ALL


def ok(text):
    """Format text as heavy blue."""
    return Fore.BLUE + Style.BRIGHT + text + Style.RESET_ALL


def success(text):
    """Format text as heavy green."""
    return Fore.GREEN + Style.BRIGHT + text + Style.RESET_ALL


class InputSpec:
    """Create an item for an InputForm."""

    def __init__(
            self, key, description, widget_spec, long_desc=None,
            validator=None, err_msg=None):
        """Initialize an input form item.

        :param key: a unique key for the item. This is used within a
            `InputForm` to retrieve the value associated with the input.
        :param description: a short description of the input.
        :param widget_spec: anything that is accepted by
            `ipywidgets.interactive`, for clarity it may be best to give
            instances of widgets rather than abbreviations.
        :param long_desc: A long description of the value.
        :param validator: Callback to validate values, should accept a single
            argument and return True/False to indicate the value is valid.
        :param err_msg: Message associated with invalid values. The string may
            contain the formatting placeholders '{value}' and '{param}', which
            will be filled with the value and self.description.
        """
        self.key = key
        self.description = description
        self.widget_spec = widget_spec
        self.long_desc = long_desc
        self.validator = validator
        if self.validator is None:
            self.validator = lambda x: True
        self.err_msg = err_msg
        if self.err_msg is None:
            self.err_msg = "Invalid value '{value}' for '{param}' parameter."

    def valid_value(self, value):
        """Return whether a given value passes validation.

        :param value: value to validate.

        :returns: (True/False, error string or None)
        """
        valid, msg = True, None
        if not self.validator(value):
            valid = False
            msg = self.err_msg.format(value=value, param=self.description)
        return valid, msg


class InputForm:
    """Easily create an input form in a notebook and store values."""

    def __init__(self, *args, widget_width='400px', description_width='150px'):
        """Initialize a form.

        :param args: args should be each an `InputSpec`.

        The current value of an item can be accessed as an attribute of the
        class using the key. The short description is displayed alongside the
        input form item, while the long description is displayed above the
        item.

        >>> inputs = InputForm(
        >>>     InputSpec(
        >>>         'a', 'My a value',
        >>>         {'min':-10, 'max':30, 'step':1, 'value':1},
        >>>         validator=lambda x: x==1,
        >>>         err_msg="Value must be equal to one."),
        >>>     InputSpec(
        >>>         'input_file', 'Input Bam', 'sample/data/mybam.bam',
        >>>         long_desc='Your input bam file should be a bam file',
        >>>         validator=lambda x: os.path.isfile(x),
        >>>         err_msg="'{value}' is not an existing file '{param}'"),
        >>>     InputSpec(
        >>>         'b', 'My b value',
        >>>         widgets.IntSlider(value=-3, max=30, min=-10)))
        >>> inputs.display()
        """
        self.widgets = dict()   # value holding widgets
        self._widgets = dict()  # for display
        self.options = args
        for option in self.options:
            if option.long_desc is not None:
                self._widgets['{}_desc'.format(option.key)] = \
                    widgets.HTML(option.long_desc)
            setter = functools.partial(self._set_value, self)
            setter.__name__ = 'setter_{}'.format(option.key)
            # create interactive widget, saves writing callback,
            # .result on the object gives the current value
            spec = {'value': option.widget_spec}
            widget = widgets.interactive(setter, **spec)
            # grab the useful part of the above and manipulate for consistency
            wid = widget.children[0]
            wid.description = ''
            wid.style.description_width = '0px'
            wid.layout = widgets.Layout(width=widget_width)
            # add label in a consistent manner
            self._widgets[option.key] = widgets.HBox([
                widgets.Label(
                    option.description,
                    layout=widgets.Layout(width=description_width)),
                widget])
            self.widgets[option.key] = widget

    def add_process_button(self, callback):
        """Add a button to do something with current values.

        :param callback: a function, should accept this class
            as a single argument. The intention here is the callback
            can mutate the class to add additional attributes after
            performing some work.
        """
        self._widgets['enter_button'] = widgets.Button(
            description='Enter', icon='angle-right')
        self._widgets['process_output'] = widgets.Output()

        def set_button_idle():
            self._widgets['enter_button'].description = 'Enter'
            self._widgets['enter_button'].disabled = False
            self._widgets['enter_button'].icon = 'angle-right'

        def set_button_busy():
            self._widgets['enter_button'].description = 'Processing'
            self._widgets['enter_button'].disabled = True
            self._widgets['enter_button'].icon = 'fa-spinner'

        def on_butt_clicked(b):
            set_button_busy()
            with self._widgets['process_output']:
                display.clear_output()
                callback(self)
                set_button_idle()

        self._widgets['enter_button'].on_click(on_butt_clicked)

    def display(self):
        """Display the input form."""
        display.display(widgets.VBox([*self._widgets.values()]))

    def validate(self):
        """Validate inputs.

        :returns: a tuple: (True/False, list of error messages)
        """
        valid = True
        errors = list()
        for desc, value, opt_valid, err_msg in self:
            if not opt_valid:
                valid = False
                errors.append(err_msg)
        return valid, errors

    def report(self, colour=False):
        """Create a report of values and errors.

        :param colour: colourize error and success text.
        """
        results = list()
        all_valid = True
        for desc, value, valid, err_msg in self:
            results.append('{}: {}'.format(desc, value))
            if not valid:
                all_valid = False
                msg = " - {}".format(err_msg)
                if colour:
                    msg = error(msg)
                results.append(msg)
        if all_valid:
            msg = "All inputs passed validation."
            if colour:
                msg = success(msg)
        else:
            msg = "One or more values failed validation."
            if colour:
                msg = error(msg)
        results.append(msg)
        return '\n'.join(results)

    def __getattr__(self, key):
        """Fetch the value of a widget item, or just a plain attribute."""
        if key in self.widgets:
            attr = self.widgets[key].result
        else:
            try:
                attr = getattr(super(), key)
            except AttributeError:
                raise AttributeError(
                    '{} instance has no attribute {}'.format(type(self), key))
        return attr

    def __iter__(self):
        """Iterate over option values.

        :yields: description, value, is valid?, error message
        """
        for option in self.options:
            value = getattr(self, option.key)
            valid, err_msg = option.valid_value(value)
            yield option.description, value, valid, err_msg

    @staticmethod
    def _set_value(self, value):
        return value



# flake8: noqa
import datetime
from dataclasses import dataclass
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, DateTime


Base = declarative_base()


@dataclass
class Statuses:
    UNKNOWN = 'UNKNOWN'
    LAUNCHED = 'LAUNCHED'
    ENCOUNTERED_ERROR = 'ENCOUNTERED_ERROR'
    COMPLETED_SUCCESSFULLY = 'COMPLETED_SUCCESSFULLY'
    TERMINATED = 'TERMINATED'


class Instance(Base):
    __tablename__ = 'instances'

    id = Column(String, primary_key=True)
    pid = Column(Integer, nullable=True)
    path = Column(String, nullable=False)
    name = Column(String, nullable=False)
    workflow = Column(String, nullable=False)
    status = Column(String, default=Statuses.UNKNOWN)
    created_at = Column(DateTime, default=datetime.datetime.utcnow)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.utcnow)

    def __init__(self, id, path, name, workflow):
        self.id = id
        self.path = path
        self.name = name
        self.workflow = workflow

    def to_dict(self):
        return {
            'id': self.id,
            'path': self.path,
            'name': self.name,
            'status': self.status,
            'workflow': self.workflow,
            'created_at': self.created_at.strftime("%Y-%m-%d-%H-%M"),
            'updated_at': self.updated_at.strftime("%Y-%m-%d-%H-%M")
        }

    def __repr__(self):
        return '<Instance %r>' % self.id


def get_session(uri):
    engine = create_engine(uri)
    Base.metadata.create_all(engine)
    Session = sessionmaker(bind=engine)
    return Session()




import os
import sys
import shutil
import re
import shutil
import platform
from glob import glob
from setuptools import setup, find_packages, Extension
from setuptools import Distribution, Command
from setuptools.command.install import install
import pkg_resources


#TODO: fill in these
__pkg_name__ = 'epi2melabs'
__author__ = 'cwright'
__description__ = 'EPI2ME labs notebook utilities.'

# Use readme as long description and say its github-flavour markdown
from os import path
this_directory = path.abspath(path.dirname(__file__))
kwargs = {'encoding':'utf-8'} if sys.version_info.major == 3 else {}
with open(path.join(this_directory, 'README.md'), **kwargs) as f:
    __long_description__ = f.read()
__long_description_content_type__ = 'text/markdown'

__path__ = os.path.dirname(__file__)
__pkg_path__ = os.path.join(os.path.join(__path__, __pkg_name__))

# Get the version number from __init__.py
verstrline = open(os.path.join(__pkg_name__, '__init__.py'), 'r').read()
vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(vsre, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in "{}/__init__.py".'.format(__pkg_name__))

dir_path = os.path.dirname(__file__)
with open(os.path.join(dir_path, 'requirements.txt')) as fh:
    install_requires = [
        str(requirement) for requirement in 
        pkg_resources.parse_requirements(fh)]

data_files = []
extra_requires = {}
extensions = []


setup(
    name=__pkg_name__,
    version=__version__,
    url='https://github.com/epi2me-labs/{}'.format(__pkg_name__),
    author=__author__,
    author_email='{}@nanoporetech.com'.format(__author__),
    description=__description__,
    long_description=__long_description__,
    long_description_content_type=__long_description_content_type__,
    dependency_links=[],
    ext_modules=extensions,
    install_requires=install_requires,
    tests_require=[].extend(install_requires),
    extras_require=extra_requires,
    # don't include any testing subpackages in dist
    packages=find_packages(exclude=['*.test', '*.test.*', 'test.*', 'test']),
    package_data={__pkg_name__:[]},
    zip_safe=False,
    data_files=data_files,
    entry_points={
        'console_scripts': [
            'cecho = {}.notebook:cecho'.format(__pkg_name__),
            'invoke_nextflow=epi2melabs.workflows.invoke:main'
        ]
    },
    scripts=[
        'scripts/concatenate_fastqs',
        'scripts/rationalise_fastq',
        'scripts/rationalize_fastq',
        'scripts/run',
        'scripts/seqkit_bam_yaml']
)



# flake8: noqa
import os
import sys
import logging
import argparse
import subprocess
from epi2melabs.workflows.database import get_session, Instance, Statuses


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="Execute a netflow workflow and update the database.",
        usage=(
            "invoke_nextflow -w epi2melabs/wf-alignment "
            "-i <instance_id> -w <workflow_name> -p <params_file> "
            "-wd <work_dir> -l <log_file> -s <stdout_file> -d <database>"
        )
    )

    parser.add_argument(
        '-n',
        '--nextflow',
        required=True,
        default='nextflow',
        help='Path to the nextflow executable.'
    )

    parser.add_argument(
        '-i',
        '--id',
        required=True,
        help='ID of the database instance record to acquire and update.'
    )

    parser.add_argument(
        '-w',
        '--workflow',
        required=True,
        help='Path to or name of the workflow to be run.'
    )

    parser.add_argument(
        '-p',
        '--params',
        required=True,
        help='Path to the workflow params file.'
    )

    parser.add_argument(
        '-wd',
        '--work_dir',
        required=True,
        help='Path to what should become the working directory.'
    )

    parser.add_argument(
        '-l',
        '--log_file',
        required=True,
        help='Path to which the logs should be written.'
    )

    parser.add_argument(
        '-s',
        '--std_out',
        required=True,
        help='Path to which the stdout should be written.'
    )

    parser.add_argument(
        '-d',
        '--database',
        required=True,
        help='Path to the SQLITE database to update.'
    )

    start = 0
    if 'invoke_nextflow' in sys.argv[0]:
        start = 1

    return parser.parse_args(sys.argv[start:])


def invoke(id: str, workflow: str, params: str, work_dir: str, 
    log_file:str, std_out: str, database: str, nextflow: str):
    logging.basicConfig(
        format='invoke_nextflow <%(asctime)s>: %(message)s',
        level=logging.DEBUG)

    logging.info('Initialising workflow.')

    # Get the command
    command = (
        f'{nextflow} -log {log_file} run {workflow} -params-file {params} '
        f'-w {work_dir} -ansi-log false')
    logging.info(f'Command: {command}.')

    # Modify if we're on windows
    if sys.platform in ["win32"]:
        logging.info("Detected OS as Windows.")
        command = 'wsl ' + command

    # Get the invocation instance by id
    db = get_session(database)
    invocation = db.query(Instance).get(id)

    # Update the invocation with the current pid
    pid = os.getpid()
    logging.info(f'The wrapper PID is {pid}.')
    invocation.pid = pid
    db.commit()

    # Set up outputs
    cli_logfile = open(std_out, 'a')
    stdout = cli_logfile
    stderr = cli_logfile
    proc = None

    try:
        # Invoke the command
        logging.info('Launching workflow.')
        proc = subprocess.Popen(
            command.split(' '), stdout=stdout, stderr=stderr)
        logging.info(f'The workflow PID is {proc.pid}.')

        # Set initial database status
        invocation.status = Statuses.LAUNCHED
        db.commit()

        # Wait for the exit status
        ret = proc.wait()
        sys.exit(ret)

    # If we receive sigint, assume the process was
    # terminated intentionally and exit gracefully
    except KeyboardInterrupt:
        logging.info('Interrupt detected: terminating workflow.')
        if proc:
            proc.kill()
        invocation.status = Statuses.TERMINATED
        db.commit()
        sys.exit(0)

    except SystemExit as e:
        # If we receive system exit of 0, assume the process
        # ended peacefully and exit gracefully.
        if not e.code:
            logging.info('Workflow completed.')
            invocation.status = Statuses.COMPLETED_SUCCESSFULLY
            db.commit()
            sys.exit(0)

        # If we receive a non-zero system exit update the
        # status to reflect an error. Exit with code 1.
        logging.info('Workflow encountered an error.')
        logging.info('See nextflow output for details.')
        invocation.status = Statuses.ENCOUNTERED_ERROR
        db.commit()
        sys.exit(1)

    # This error is thrown if the path to Nextflow
    # is not available, and therefore cannot be launched
    except FileNotFoundError as e:
        logging.info(f"Cant find '{nextflow}' on the path.")
        logging.info(e)
        invocation.status = Statuses.ENCOUNTERED_ERROR
        db.commit()
        sys.exit(1)

    # Handle all other exception classes in the event of
    # unhandled exceptions occurring within the callable.
    # Set the status to error and exit with code 1.
    except Exception as e:
        logging.info('Workflow encountered an error.')
        logging.info(e)
        invocation.status = Statuses.ENCOUNTERED_ERROR
        db.commit()
        sys.exit(1)


def main():
    args = parse_args()
    invoke(
        id=args.id,
        workflow=args.workflow,
        params=args.params,
        work_dir=args.work_dir,
        log_file=args.log_file,
        std_out=args.std_out,
        database=args.database,
        nextflow=args.nextflow)


if __name__ == '__main__':
    main()


# flake8: noqa
import re
import os
import json
import signal
import shutil
import requests
import shortuuid
import subprocess
from datetime import datetime
from dataclasses import dataclass
from urllib.parse import quote_plus
from nf_core.schema import PipelineSchema
from jsonschema.validators import validator_for
from typing import Callable, Dict, Type, TypedDict, Union, Tuple, List, Any
from epi2melabs.workflows.database import Statuses, get_session, Instance


class LauncherException(Exception):
    ...


class Path(TypedDict):
    name: str
    path: str
    isdir: bool
    updated: float

class PathResponse(TypedDict):
    path: str
    exists: bool
    error: Union[str, None]

class FilePathResponse(PathResponse):
    contents: Union[None, List[str]]

class DirectoryPathResponse(PathResponse):
    contents: Union[None, List[Path]]

class InstanceCreateResponse(TypedDict):
    created: bool
    instance: dict
    error: Union[str, None]

@dataclass
class Workflow:
    url: str
    name: str
    desc: str
    path: str
    target: str
    schema: Dict
    defaults: Dict

    def to_dict(self):
        return {
            'url': self.url,
            'name': self.name,
            'desc': self.desc,
            'path': self.path,
            'target': self.target,
            'schema': self.schema,
            'defaults': self.defaults}


class WorkflowLauncher(object):
    MAINNF: str = 'main.nf'
    PARAMS: str = 'params.json'
    SCHEMA: str = 'nextflow_schema.json'
    
    def __init__(self, base_dir=None, workflows_dir=None,
        invoker='invoke_nextflow', nextflow='nextflow') -> None:

        self._workflows: Dict[str, Workflow] = {}

        self.curr_dir = os.getcwd()
        self.base_dir = base_dir or os.path.join(os.getcwd(), 'nextflow')
        self.workflows_dir = workflows_dir or os.path.join(self.base_dir, 'workflows')
        self.instance_dir = os.path.join(self.base_dir, 'instances')
        self.database_uri = f'sqlite:///{self.base_dir}/db.sqlite'

        self.invoker = invoker
        self.nextflow = nextflow

        self.get_or_create_dir(self.base_dir)
        self.get_or_create_dir(self.instance_dir)
        self.get_or_create_dir(self.workflows_dir)
        self.db = get_session(self.database_uri)

    #
    # Workflows
    #
    @property
    def workflows(self) -> Dict:
        for item in os.listdir(self.workflows_dir):
            if self._workflows.get(item):
                continue
            path = os.path.join(self.workflows_dir, item)
            if not os.path.isdir(path):
                continue
            if not self.SCHEMA in os.listdir(path):
                continue
            try:
                self._workflows[item] = self.load_workflow(
                    item, path)
            except LauncherException:
                pass
        return {
            workflow.name: workflow.to_dict()
            for workflow in self._workflows.values()
        }

    def get_workflow(self, name: str) -> Dict:
        if workflow := self.workflows.get(name):
            return workflow
        return {}

    def _get_workflow(self, name: str) -> Union[Workflow, None]:
        if not self.workflows:
            pass
        if workflow := self._workflows.get(name):
            return workflow
        return None

    def load_workflow(self, name: str, path: str) -> Workflow:
        schema = self._load_schema(os.path.join(path, self.SCHEMA))
        defaults = self._get_schema_defaults(schema)

        if not schema:
            raise LauncherException(
                f'Cannot reload {name}: missing schema')

        target = None
        main_nf = os.path.join(path, self.MAINNF)
        if os.path.exists(main_nf):
            target = main_nf

        if not target:
            target = schema.get('title', ' ')
            if ' ' in target:
                raise LauncherException(
                    'Cannot load {name}: invalid title')

        if not target:
            raise LauncherException(
                'Cannot load {name}: no entrypoint')

        return Workflow(
            name = name,
            path = path,
            target = target,
            schema = schema,
            defaults = defaults,
            url =  schema.get('url', '#'),
            desc = schema.get('description', ''))

    #
    # Schema
    #
    def _load_schema(self, path: str) -> Dict:
        with open(path, "r") as schema:
            return json.load(schema)

    def _get_schema_defaults(
        self, schema: Dict
    ) -> Dict[str,  Union[str, int, float, bool]]:
        return {
            **self._get_schema_defaults_for_fragment(
                schema.get("properties", {})),
            **self._get_schema_defaults_for_fragment(
                schema.get("definitions", {}).get("properties", {}))}

    def _get_schema_defaults_for_fragment(
        self, fragment: Dict
    ) -> Dict[str, Union[str, int, float, bool]]:
        defaults = {}
        for key, param in fragment.items():
            default = self._get_parameter_default(param)
            if default is None:
                continue
            defaults[key] = default
        return defaults

    def _get_parameter_default(
        self, param: Dict
    ) -> Union[str, int, float, bool, None]:
        default = None
        if "default" in param:
            default = self._sanitise_parameter_default(
                param['default'], param.get('type'))
        return default

    def _sanitise_parameter_default(self,
        default: Union[str, int, float, bool], _type=None
    ) -> Union[str, int, float, bool]:
        if not _type:
            return default
        elif _type == "boolean":
            if not isinstance(default, bool):
                default = (default == "true")
            return default
        elif isinstance(default, str) and default.strip() == "":
            return ""
        elif _type == "integer":
            return int(default)
        elif _type == "number":
            return float(default)
        return str(default)

    #
    # Instances
    #
    @property
    def instances(self) -> Dict:
        return {
            instance.id: instance.to_dict()
            for instance in self.db.query(Instance).all()
        }

    def get_instance(self, id: str) -> Dict:
        if instance := self._get_instance(id):
            return instance.to_dict()
        return {}

    def _get_instance(self, id: str) -> Union[Instance, None]:
        if instance := self.db.query(Instance).get(id):
            return instance
        return None

    def create_instance(
        self, name: str, workflow_name: str, params: Dict
    ) -> InstanceCreateResponse:
        workflow = self._get_workflow(workflow_name)

        if not workflow or not workflow.target:
            return InstanceCreateResponse(
                created=False, instance={},
                error='Could not create instance, workflow unknown')

        return self._create_instance(
            name, workflow.name, workflow.target, params)

    def _create_instance(
        self, name: str, workflow_name: str, workflow_target: str, params: Dict
    ) -> InstanceCreateResponse:

        # Generate instance details
        _id = str(shortuuid.uuid())
        now = datetime.now().strftime("%Y-%m-%d-%H-%M")
        dirname = '_'.join([now, workflow_name, _id])
        path = os.path.join(self.instance_dir, dirname)

        # Create instance db record
        instance = Instance(_id, path, name, workflow_name)
        self.db.add(instance)
        self.db.commit()

        # Construct instance filepaths
        params_file = os.path.join(path, self.PARAMS)
        nf_logfile = os.path.join(path, 'nextflow.log')
        nf_std_out = os.path.join(path, 'nextflow.stdout')
        iv_std_out = os.path.join(path, 'invoke.stdout')
        out_dir = os.path.join(path, 'output')
        work_dir = os.path.join(path, 'work')

        # Touch instance files and directories
        self.get_or_create_dir(path)
        self.get_or_create_dir(out_dir)
        self.get_or_create_dir(work_dir)

        for targ in [nf_logfile, nf_std_out, iv_std_out]:
            with open(targ, 'a'):
                pass

        # Coerce params accordingly
        params = self._fix_parameters(workflow_name, **params)
        params['out_dir'] = out_dir # Todo: generalise to support 3rd party wfs

        with open(params_file, 'w') as json_file:
            json_file.write(json.dumps(params, indent=4))

        # Launch process
        error = self._start_instance(instance, workflow_target, params_file,
                work_dir, nf_logfile, nf_std_out, iv_std_out, self.database_uri)

        return InstanceCreateResponse(
            created=True, instance=instance.to_dict(), error=error)

    def _start_instance(self, instance, target: str, params_file: str,
        work_dir: str, nf_logfile: str, nf_std_out: str,
        iv_std_out: str, database: str) -> Union[str, None]:

        logfile = open(iv_std_out, 'a')
        stdout = logfile
        stderr = logfile

        command = (
            f'{self.invoker} -i {instance.id} -w {target} -p {params_file} '
            f'-wd {work_dir} -l {nf_logfile} -s {nf_std_out} -d {database} '
            f'-n {self.nextflow}')

        # Ensure the default location for docker on MacOS is available
        env = os.environ.copy()
        env["PATH"] = "/usr/local/bin:" + env["PATH"]

        # Todo: May need to provide some different flags on windows
        # in lieu of start_new_session working
        try:
            proc = subprocess.Popen(command.split(' '), start_new_session=True,
                stdout=stdout, stderr=stderr, close_fds=True, cwd=self.base_dir,
                env=env)
        except FileNotFoundError:
            error = 'Could not launch instance, invocation script not found'
            logfile.write(error)
            instance.status = Statuses.ENCOUNTERED_ERROR
            self.db.commit()
            return error

        instance.pid = proc.pid
        self.db.commit()

    def delete_instance(self, id: str, delete: bool = False) -> bool:
        instance = self._get_instance(id)
        if not instance:
            return False

        # Stop any process
        self._stop_instance(instance)

        # Optionally delete the directory
        if delete:
            try:
                shutil.rmtree(instance.path)
            except FileNotFoundError:
                pass

            # Delete record
            self.db.delete(instance)
            self.db.commit()

        return True

    def _stop_instance(self, instance) -> bool:
        if instance.status != Statuses.LAUNCHED:
            return False

        try:
            os.kill(int(instance.pid), signal.SIGINT)
            return True
        except (OSError, KeyboardInterrupt, TypeError):
            pass

        return False

    #
    # Pathing
    #
    def get_path(self, path: str) -> PathResponse:
        path = self._fix_path(path)
        exists, error = self._check_path(path)
        return PathResponse(path=path, exists=exists, error=error)

    def get_file(self, path: str, contents: bool = False) -> FilePathResponse:
        lines = None
        path = self._fix_path(path)
        exists, error = self._check_path(path)
        if exists and not os.path.isfile(path):
            error = 'Path is not a file'
            return FilePathResponse(
                path=path, exists=exists, error=error, contents=None)
        if exists and contents:
            lines = self.read_file(path)
        return FilePathResponse(
            path=path, exists=exists, error=error, contents=lines)

    def read_file(self, path: str) -> List[str]:
        lines = []
        with open(path) as lf:
            lines = lf.readlines()
            lines = [self._process_line(line) for line in lines]
        return lines

    def _process_line(self, line: str) -> str:
        line = line.rstrip()
        # 7-bit C1 ANSI sequences
        # https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        return ansi_escape.sub('', line)

    def get_directory(self, path: str, contents: bool = False) -> DirectoryPathResponse:
        items = None
        path = self._fix_path(path)
        exists, error = self._check_path(path)
        if exists and not os.path.isdir(path):
            error = 'Path is not a directory'
            return DirectoryPathResponse(
                path=path, exists=exists, error=error, contents=None)
        if exists and contents:
            items = self.list_dir(path)
        return DirectoryPathResponse(
            path=path, exists=exists, error=error, contents=items)

    def list_dir(self, path: str) -> List[Path]:
        items = []
        abspath = os.path.abspath(path)
        for item in os.listdir(path):
            if item.startswith('.'):
                continue
            item_abspath = os.path.join(abspath, item)
            modified = os.path.getmtime(item_abspath)
            is_dir = os.path.isdir(item_abspath)
            items.append(Path(
                name=item,
                path=item_abspath,
                updated=modified,
                isdir=is_dir))
        return items

    def _fix_path(self, path: str) -> str:
        if not os.path.abspath(path):
            path = os.path.join(self.base_dir, path)
        return path

    def _check_path(self, path) -> Tuple[bool, str]:
        if os.path.exists(path):
            return True, 'Path exists.'
        return False, 'Path does not exist or host cannot see it.'

    def get_or_create_dir(self, path: str) -> Tuple[bool, str]:
        if not os.path.exists(path):
            os.mkdir(path)
            return True, os.path.abspath(path)
        return False, os.path.abspath(path)

    def _fix_parameters(self, workflow_name, **params):
        coerced = {}
        for param_key, param_value in params.items():
            schema = self._get_schema_for_param(workflow_name, param_key)
            fmt = schema.get('format')
            if fmt in ['path', 'file-path', 'directory-path']:
                path = param_value
                if not os.path.isabs(path):
                    path = os.path.join(self.curr_dir, path)
                coerced[param_key] = path
                continue
            coerced[param_key] = param_value
        return coerced

    def _get_schema_for_param(self, workflow_name: str, param_name: str) -> Dict:
        if workflow := self._get_workflow(workflow_name):
            sections = workflow.schema.get('definitions', {})
            for section in sections.values():
                for k, v in section.get('properties', {}).items():
                    if k == param_name:
                        return v
        return {}


class RemoteWorkflowLauncher(WorkflowLauncher):

    def __init__(self, base_dir, workflows_dir,
        ip: str = '0.0.0.0', port: str = '8090'):
        self.ip = ip
        self.port = port

        super().__init__(base_dir, workflows_dir)

    @property
    def instances(self) -> Dict:
        response = requests.get(
            f'http://{self.ip}:{self.port}/instances?uuid={shortuuid.uuid()}')
        return response.json()

    def get_instance(self, id: str) -> Dict:
        response = requests.get(
            f'http://{self.ip}:{self.port}/instances/{id}?uuid={shortuuid.uuid()}')
        return response.json()

    def create_instance(
        self, name: str, workflow_name: str, params: Dict
    ) -> InstanceCreateResponse:
        workflow = self._get_workflow(workflow_name)

        if not workflow or not workflow.target:
            return InstanceCreateResponse(
                created=False, instance={},
                error='Could not create instance, workflow unknown')

        response = requests.post(
            f'http://{self.ip}:{self.port}/instances',
            data=json.dumps({
                'name': name,
                'workflow_name': workflow_name,
                'workflow_target': workflow.target,
                'params': params,
            }),
            headers={
                'Content-type': 'application/json', 
                'Accept': 'text/plain'
            })

        data = response.json()
        return data

    def delete_instance(self, id: str, delete: bool = False) -> bool:
        response = requests.delete(
            f'http://{self.ip}:{self.port}/instances/{id}?delete={delete}')

        data = response.json()
        return data['deleted']

    def get_path(self, path: str) -> PathResponse:
        response = requests.get(
            f'http://{self.ip}:{self.port}/path/{quote_plus(quote_plus(path))}')
        return response.json()

    def get_file(self, path: str, contents: bool = False) -> FilePathResponse:
        _cont = ''
        if contents:
            _cont = f'?contents={shortuuid.uuid()}'
        response = requests.get(
            f'http://{self.ip}:{self.port}/file/{quote_plus(quote_plus(path))}{_cont}')
        return response.json()

    def get_directory(self, path: str, contents: bool = False) -> DirectoryPathResponse:
        _cont = ''
        if contents:
            _cont = f'?contents={shortuuid.uuid()}'
        response = requests.get(
            f'http://{self.ip}:{self.port}/directory/{quote_plus(quote_plus(path))}{_cont}')
        return response.json()


def get_workflow_launcher(
    base_dir, workflows_dir, remote=False, ip=None, port=None):

    if remote:
        kwargs = {'ip': ip, 'port': port}
        for k, v in kwargs.items():
            if v is None:
                kwargs.pop(k)
        return RemoteWorkflowLauncher(base_dir, workflows_dir, **kwargs)

    return WorkflowLauncher(base_dir, workflows_dir)